
/*
 * $Id: database.c,v 1.32 2005/10/26 19:29:16 jheinonen Exp $
 *
 * by JH <jheinonen@users.sourceforge.net>
 *
 * Copyright (C) Jaakko Heinonen
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>
#ifdef HAVE_CONFIG_H
#	include "config.h"
#endif
#include "abook.h"
#include "database.h"
#include "filter.h"
#include "filters/filter_misc.h" /* get_surname() */
#include "gettext.h"
#include "list.h"
#include "misc.h"
#include "xmalloc.h"


abook_field_list *fields_list;

list_item *database = NULL;
int items = 0;

#define ITEM_SIZE (fields_list->count * sizeof(char *))
#define LAST_ITEM (items - 1)

#define INITIAL_LIST_CAPACITY	30
static int list_capacity = 0;

int standard_fields_indexed[ITEM_FIELDS];

/*
 * notes about adding predefined "standard" fields:
 * 	- leave alone "name" and "email"
 * 	- reorganize the field numbers in database.h
 */
abook_field standard_fields[] = {
	{"name",	N_("Name"),		FIELD_STRING}, /* NAME */
	{"email",	N_("E-mail addresses"),	FIELD_EMAILS}, /* EMAIL */
	{"address",	N_("Address"),		FIELD_STRING}, /* ADDRESS */
	{"address2",	N_("Address2"),		FIELD_STRING}, /* ADDRESS2 */
	{"city",	N_("City"),		FIELD_STRING}, /* CITY */
	{"state",	N_("State/Province"),	FIELD_STRING}, /* STATE */
	{"zip",		N_("ZIP/Postal Code"),	FIELD_STRING}, /* ZIP */
	{"country",	N_("Country"),		FIELD_STRING}, /* COUNTRY */
	{"phone",	N_("Home Phone"),	FIELD_STRING}, /* PHONE */
	{"workphone",	N_("Work Phone"),	FIELD_STRING}, /* WORKPHONE */
	{"fax",		N_("Fax"),		FIELD_STRING}, /* FAX */
	{"mobile",	N_("Mobile"),		FIELD_STRING}, /* MOBILEPHONE */
	{"nick",	N_("Nickname/Alias"),	FIELD_STRING}, /* NICK */
	{"url",		N_("URL"),		FIELD_STRING}, /* URL */
	{"notes",	N_("Notes"),		FIELD_STRING}, /* NOTES */
	{"anniversary",	N_("Anniversary day"),	FIELD_DAY},    /* ANNIVERSARY */
	{0} /* ITEM_FIELDS */
};


extern int first_list_item;
extern int curitem;
extern char *selected;
extern char *datafile;


static abook_field *
declare_standard_field(int i)
{
	abook_field *f = xmalloc(sizeof(abook_field));

	f = memcpy(f, &standard_fields[i], sizeof(abook_field));
	f->name = xstrdup(gettext(f->name));

	add_field(fields_list, f);

	assert(standard_fields_indexed[i] == -1);
	standard_fields_indexed[i] = fields_list->count - 1;

	return f;
}

abook_field *
find_standard_field(char *key, int do_declare)
{
	int i;

	for(i = 0; standard_fields[i].key; i++)
		if(0 == strcmp(standard_fields[i].key, key))
			goto found;

	return NULL;

found:
	return do_declare ? declare_standard_field(i) : &standard_fields[i];
}

/* Search for a field. Use the list of declared fields if no list specified. */
abook_field *
real_find_field(char *key, abook_field_list *list, int *number)
{
	int i;

	for(i = 0, list = list ? list : fields_list; i < list->count; i++)
		if(0 == strcmp(list->fields[i]->key, key)) {
			if(number)
				*number = i;
			return list->fields[i];
		}

	if(number)
		*number = -1;

	return NULL;
}

void
add_field(abook_field_list *list, abook_field *f)
{
	int i = ++list->count;

	list->fields = xrealloc(list->fields, i * sizeof(abook_field));
	list->fields[i - 1] = f;
}

char *
declare_new_field(char *key, char *name, char *type, int accept_standard)
{
	abook_field *f;

	if(find_declared_field(key))
		return _("field already defined");

	if(find_standard_field(key, accept_standard))
		return accept_standard ? NULL /* ok, added */ :
			_("standard field does not need to be declared");

	f = xmalloc(sizeof(abook_field));
	f->key = xstrdup(key);
	f->name = xstrdup(name);

	if(!*type || (0 == strcasecmp("string", type)))
		f->type = FIELD_STRING;
	else if(0 == strcasecmp("emails", type))
		f->type = FIELD_EMAILS;
	else if(0 == strcasecmp("list", type))
		f->type = FIELD_LIST;
	else if(0 == strcasecmp("day", type))
		f->type = FIELD_DAY;
	else
		return _("unknown type");

	add_field(fields_list, f);

	return NULL;
}

/*
 * Declare "standard" fields, thus preserving them while parsing a database,
 * even if they won't be displayed.
 */
void
init_standard_fields()
{
	int i;

	for(i = 0; standard_fields[i].key; i++)
		if(standard_fields_indexed[i] == -1)
			declare_standard_field(i);
}

/* Some initializations - Must be called _before_ load_opts() */
void
prepare_database_internals()
{
	int i;

	fields_list = xmalloc0(sizeof(abook_field_list));

	for(i = 0; i < ITEM_FIELDS; i++)
		standard_fields_indexed[i] = -1;

	/* the only two mandatory fields */
	declare_standard_field(NAME);
	declare_standard_field(EMAIL);
}

int
load_database(char *filename)
{
	if(database != NULL)
		close_database();

	if(import_file("abook", filename) != 0)
		return -1;

	return (items == 0) ? 2 : 0;
}

int
save_database()
{
	FILE *out;
	int ret = 0;
	char *datafile_new = strconcat(datafile, ".new", NULL);
	char *datafile_old = strconcat(datafile, "~", NULL);

	if( (out = abook_fopen(datafile_new, "w")) == NULL ) {
		ret = -1;
		goto out;
	}

	if(list_is_empty()) {
		fclose(out);
		unlink(datafile);
		ret = 1;
		goto out;
	}

	ret = fexport("abook", out, ENUM_ALL);
	fclose(out);

	if(access(datafile, F_OK) == 0 &&
			(rename(datafile, datafile_old)) == -1)
		ret = -1;
	
	if((rename(datafile_new, datafile)) == -1)
		ret = -1;

out:
	free(datafile_new);
	free(datafile_old);
	return ret;
}

static void
db_free_item(int item)
{
	item_empty(database[item]);
}

void
close_database()
{
	int i;

	for(i=0; i <= LAST_ITEM; i++)
		db_free_item(i);

	xfree(database);
	free(selected);

	database = NULL;
	selected = NULL;

	items = 0;
	first_list_item = curitem = -1;
	list_capacity = 0;
}

void
adjust_list_capacity()
{
	if(items <= list_capacity)
		return;

	if(list_capacity < 1)
		list_capacity = INITIAL_LIST_CAPACITY;
	else if(items >= list_capacity)
		list_capacity *= 2;
	else if(list_capacity / 2 > items)
		list_capacity /= 2;
	else
		return;

	if(database)
		database = xrealloc(database,sizeof(list_item) * list_capacity);
	else /* allocate memory _and_ initialize pointers to NULL */
		database = xmalloc0(sizeof(list_item) * list_capacity);

	selected = xrealloc(selected, list_capacity);
}

void
remove_selected_items()
{
	int i, j;

	if(list_is_empty())
		return;

	if(!selected_items())
		selected[curitem] = 1;

	for(j = LAST_ITEM; j >= 0; j--) {
		if(selected[j]) {
			db_free_item(j); /* added for .4 data_s_ */
			for(i = j; i < LAST_ITEM; i++) {
				item_copy(database[i], database[i + 1]);
				selected[i] = selected[i + 1];
			}
			item_free(&database[LAST_ITEM]);
			items--;
		}
	}

	if(curitem > LAST_ITEM && items > 0)
		curitem = LAST_ITEM;

	adjust_list_capacity();

	select_none();
}

static int
surnamecmp(const void *i1, const void *i2)
{
	int ret, idx = field_id(NAME);
	char *n1, *n2, *s1, *s2;

	n1 = (*(list_item *)i1)[idx];
	n2 = (*(list_item *)i2)[idx];
	
	s1 = get_surname(n1);
	s2 = get_surname(n2);

	if( !(ret = safe_strcoll(s1, s2)) )
		ret = safe_strcoll(n1, n2);

	free(s1);
	free(s2);

	return ret;
}

static int sort_field = -1;

static int
namecmp(const void *i1, const void *i2)
{
	char *n1, *n2;

	assert(sort_field >= 0 && sort_field < fields_list->count);

	n1 = (*(list_item *)i1)[sort_field];
	n2 = (*(list_item *)i2)[sort_field];

	return safe_strcoll(n1, n2);
}

void
sort_by_field(char *name)
{
	int field;

	select_none();

	name = (name == NULL) ? opt_get_str(STR_SORT_FIELD) : name;
	find_field_number(name, &field);

	if(field < 0) {
		if(name == opt_get_str(STR_SORT_FIELD))
			statusline_msg(_("Invalid field value defined "
				"in configuration"));
		else
			statusline_msg(_("Invalid field value for sorting"));

		return;
	}

	sort_field = field;

	qsort((void *)database, items, sizeof(list_item), namecmp);

	refresh_screen();
}

void
sort_surname()
{
	select_none();

	qsort((void *)database, items, sizeof(list_item), surnamecmp);

	refresh_screen();
}

/* TODO implement a search based on more sophisticated patterns */
int
find_item(char *str, int start, int search_fields[])
{
	int i, id;
	char *findstr = NULL;
	char *tmp = NULL;
	int ret = -1; /* not found */
	struct db_enumerator e = init_db_enumerator(ENUM_ALL);

	if(list_is_empty() || !is_valid_item(start))
		return -2; /* error */

	findstr = xstrdup(str);
	findstr = strlower(findstr);

	e.item = start - 1; /* must be "real start" - 1 */
	db_enumerate_items(e) {
		for(i = 0; search_fields[i] >= 0; i++) {
			if((id = field_id(search_fields[i])) == -1)
				continue;
			if(database[e.item][id] == NULL)
				continue;
			tmp = xstrdup(database[e.item][id]);
			if( tmp && strstr(strlower(tmp), findstr) ) {
				ret = e.item;
				goto out;
			}
			xfree(tmp);
		}
	}

out:
	free(findstr);
	free(tmp);
	return ret;
}

int
is_selected(int item)
{
	return selected[item];
}

int
is_valid_item(int item)
{
	return item <= LAST_ITEM && item >= 0;
}

list_item
db_item_get(int i)
{
	return database[i];
}

int
last_item()
{
	return items - 1;
}

