
/*
 * $Id$
 *
 * by Cedric Duval <cedricduval@free.fr>
 *
 * Copyright (C) Cedric Duval
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#ifdef HAVE_CONFIG_H
#	include "../config.h"
#endif
#include "../abook.h"
#include "../db_access.h"
#include "../misc.h"
#include "../xmalloc.h"
#include "assert.h"

#include "filter_interface.h"
#include "filter.h"

#include "external.h"

static char *i_path = NULL, *e_path = NULL;
static char *filter_desc = NULL, *alias_of = NULL;

static FILE *in_cmd, *out_data, *in_data;
static list_item item;
static int send_errors;
static struct db_enumerator db_enum;


static int cmd_field(abook_buffer *);
static int cmd_get_emails(abook_buffer *);


/*
 * Important: keep this array in sync with the corresponding enum in external.h
 */
static struct {
	char *cmd;
	int access;
	int (*func) (abook_buffer *);
} cmds[] = {
	{ "toggle_errors", ACCESS_BOTH,	NULL }, /* CMD_TOGGLE_ERRORS */
	{ "stop",	ACCESS_BOTH,	NULL }, /* CMD_STOP */
	{ "next_line",	ACCESS_IN,	NULL }, /* CMD_NEXT_LINE */
	{ "new_item",	ACCESS_IN,	NULL }, /* CMD_NEW_ITEM */
	{ "field",	ACCESS_IN,	cmd_field }, /* CMD_FIELD */
	{ "commit_item", ACCESS_IN,	NULL }, /* CMD_COMMIT_ITEM */
	{ "next_item",	ACCESS_OUT,	NULL }, /* CMD_NEXT_ITEM */
	{ "get_emails",	ACCESS_OUT,	cmd_get_emails }, /* CMD_GET_EMAILS */
	{ NULL }
};

/*
 * Important:
 *   - keep this array in sync with the corresponding enum in external.h
 *   - only append OK or error codes, in order to preserve compatibility
 */
static char *errors[] = {
	"OK", /* ERR_OK */
	"OK - new field declared", /* ERR_OK_DECLARED_NEW_FIELD */
	/* ERR_OK_SEPARATOR - error numbers starts from 1 with next entry */
	"field undeclared", /* ERR_UNDECLARED_FIELD */
	"malformed request", /* ERR_MALFORMED_REQUEST */
	"field name is mandatory", /* ERR_NO_NAME_FIELD */
	"command not authorized in this mode", /* ERR_CMD_NOT_AUTHORIZED */
};

static int send_error(int err)
{
	int i = (err < ERR_OK_SEPARATOR) ? err : err - 1;
	err = (err < ERR_OK_SEPARATOR) ? 0 : err - ERR_OK_SEPARATOR;
	assert(i >= 0 && i < sizeof(errors) / sizeof(char *));
	if(send_errors) {
		fprintf(out_data, "%d >>> %s\n", err, errors[i]);
		fflush(out_data);
	}

	return -err;
}

/*
 * Launching the external filter
 */
static int
external_run(char *path, FILE *out)
{
	int fd_command[2], fd_data[2];
	pid_t child;

	if((pipe(fd_command) == -1) || (pipe(fd_data) == -1) ||
			((child = fork()) == -1))
		return -1;

	if(child == -1) {
		fprintf(stderr, "fork failed");
		return -1;
	} if(child == 0) {
		close(fd_data[1]);
		close(fd_command[0]);

		close(STDERR_FILENO);
		dup2(fd_command[1], STDERR_FILENO);
		close(fd_command[1]);

		if(out) {
			close(STDOUT_FILENO);
			dup2(fileno(out), STDOUT_FILENO);
			close(fileno(out));
		}

		close(STDIN_FILENO);
		dup2(fd_data[0], STDIN_FILENO);
		close(fd_data[0]);

		execlp(path, path, NULL);

		exit(1);
	}

	in_cmd = fdopen(fd_command[0], "r");
	out_data = fdopen(fd_data[1], "w");
	close(fd_data[0]);
	close(fd_command[1]);

	return child;
}

/*
 * Filling a field
 * 
 * syntax: field <name> = <value>
 */
static int
cmd_field(abook_buffer *b)
{
	int field, ret;

	if(get_token(b, TOKEN_EQUAL) || !b->data)
		return send_error(ERR_MALFORMED_REQUEST);

	if((field = field_index(b->data)) != -1) {
		if(item[field])
			free(item[field]);
		if(*b->ptr)
			item[field] = xstrdup(b->ptr);
		ret = ERR_OK;
	} else if(!strcasecmp(opt_get_str(STR_PRESERVE_FIELDS),	"all")) {
		if(!*b->ptr)
			return send_error(ERR_MALFORMED_REQUEST);
		field = declare_unknown_field(b->data);
		assert(field != -1);
		item_resize(&item);
		item[field] = xstrdup(b->ptr);
		ret = ERR_OK_DECLARED_NEW_FIELD;
	} else
		ret = ERR_UNDECLARED_FIELD;

	return send_error(ret);
}

/*
 * Fetching all email addresses of current item
 */
static int
cmd_get_emails(abook_buffer *b)
{
	struct field_enumerator fe = init_field_enumerator(ENUM_EMAILS);
	abook_list *emails, *cur;

	enumerate_fields(fe) {
		emails = csv_to_abook_list(db_fget_byid(db_enum.item,fe.index));
		if(emails) {
			for(cur = emails; cur; cur = cur->next)
				send_data(cur->data);
			abook_list_free(&emails);
		}
	}
	send_data(""); /* void string indicates end of list */

	return 0;
}

/*
 * Handling of the external process requests
 */

static int
get_command(abook_buffer *b)
{
	int i;

	if(get_token(b, 0) || !b->data)
		return -1;

	for(i = 0; cmds[i].cmd; i ++)
		if(!strcasecmp(b->data, cmds[i].cmd)) {
			if((in_data && (cmds[i].access == ACCESS_OUT)) ||
				(!in_data && (cmds[i].access == ACCESS_IN)))
				return -2; /* command not authorized */

			if(cmds[i].func) {
				cmds[i].func(b);
				return -1; /* processed */
			} else
				return i; /* still to be processed */
		}
	return -3;
}

static int
process_command(abook_buffer *b)
{
	char *str;
	int cmd;

	switch(cmd = get_command(b)) {
		case CMD_NEXT_LINE:
			if(feof(in_data) || (str = getaline(in_data)) == NULL)
				return 1;
			send_data(str);
			break;
		case CMD_NEW_ITEM:
			item_empty(item);
			send_error(ERR_OK);
			break;
		case CMD_TOGGLE_ERRORS:
			send_errors = !send_errors;
			send_error(ERR_OK);
			break;
		case CMD_COMMIT_ITEM:
			if(!item[field_id(NAME)])
				send_error(ERR_NO_NAME_FIELD);
			else {
				add_item2database(item);
				send_error(ERR_OK);
			}
			item_blank(item);
			break;
		case CMD_NEXT_ITEM:
			if(db_next_item(db_enum) == -1)
				return 1;
			send_error(ERR_OK);
			break;
		case CMD_STOP:
			if(*b->ptr)
				fprintf(stderr, "%s\n", b->ptr);
			return 1;
		case -2:
			send_error(ERR_CMD_NOT_AUTHORIZED);
			break;
		default:
			assert(cmd <= 0); /*otherwise, unhandled known command*/
	}
	return 0;
}

static void
process_external_commands()
{
	char *line = NULL;
	abook_buffer b;

	item = item_create();

	while(1) {
		if(line)
			free(line);
		line = b.ptr = getaline(in_cmd);
		strtrim(b.ptr);
		if(feof(in_cmd))
			break;

		if(process_command(&b))
			break;
	}

	free(line);
	item_empty(item);
	item_free(&item);

	fclose(in_cmd);
	fclose(out_data);
}

/*
 * Export filter
 */
static int
external_export_database(FILE *out, struct db_enumerator e)
{
	pid_t child;
	if(!e_path)
		return 1;

	child = external_run(e_path, out);

	if(child < 0)
		return 1;

	db_enum = e;
	if(db_next_item(db_enum) == -1) /* advancing to first item */
		return 1;

	process_external_commands();

	waitpid(child, NULL, 0);

	return 0;
}

/*
 * Import filter
 */
static int
external_parse_file(FILE *in)
{
	pid_t child;

	if(!i_path)
		return 1;

	child = external_run(i_path, NULL);

	if(child < 0)
		return 1;

	in_data = in;

	process_external_commands();

	waitpid(child, NULL, 0);
	
	return 0;
}

/*
 * Some initializations
 */
static void
external_filter_init(abook_list *args)
{
	abook_list *a;

	for(a = args; a; a = a->next->next) {
		if(!strncmp(a->data, "description", 11))
			filter_desc = xstrdup(a->next->data);
		else if(!strncmp(a->data, "import_path", 11))
			i_path = xstrdup(a->next->data);
		else if(!strncmp(a->data, "export_path", 11))
			e_path = xstrdup(a->next->data);
		else if(!strncmp(a->data, "alias_of", 8))
			alias_of = xstrdup(a->next->data);
		else
			assert(0);
	}

	send_errors = 1; /* verbose by default */
}

static char *
external_filter_description()
{
	return safe_str(filter_desc);
}


/*
 * Using filter helper
 */
#define FILTER_NAME external
#define FILTER_DESCRIPTION N_("[call to external filter]")
#define FILTER_EXPORT_FUNCTION external_export_database
#define FILTER_IMPORT_FUNCTION external_parse_file
#define FILTER_INIT_HOOK external_filter_init
#define FILTER_DESC_HOOK external_filter_description
#include "filter_helper.h"
